"""base class and casting descriptor for tweettr"""
__author__ = "Andreas Poehlmann"
__email__ = "andreas@poehlmann.io"

import re
from typing import Any, Dict, Literal, Union, get_args, get_origin, get_type_hints
# noinspection PyUnresolvedReferences
from typing import ForwardRef

__all__ = ['TweettrBase', 'enhance_type_hinted_classes_with_descriptors']


class TweettrDescriptor:
    """non-data descriptor to transform twitter 'objects' to python instances

    Allows attribute access through all defined levels of the tweet json
    """
    def __init__(self, name, type_cls, optional=False, as_list=False):
        self._name = name
        self._type_cls = type_cls
        self._optional = optional
        self._as_list = as_list

    def __get__(self, instance, owner):
        try:
            # this accesses the underlying dict of a tweettr instance directly
            values = instance[self._name]
        except KeyError:
            if self._optional:
                return None
            elif self._as_list:
                return []
            else:
                # the requested attribute is missing in the tweet json
                # this probably means you should label this attribute
                # as 'optional' in the relevant class
                raise AttributeError(f'"{owner.__name__}.{self._name}" not found')

        if self._optional and values is None:
            return None
        if self._as_list:
            return [self._type_cls(v) for v in values]

        return self._type_cls(values)


def _typing_to_np_doc(type_, level=0):
    """convert a type decoration used in Tweettr to np doc style type"""
    optional = False
    origin = get_origin(type_)
    t_args = get_args(type_)

    if origin is Union and type(None) in t_args:
        if level == 0:
            optional = True
        output = _typing_to_np_doc(t_args[0], level=level+1)
    elif origin is list:
        output = f'List[{_typing_to_np_doc(t_args[0], level=level+1)!s}]'
    elif origin is dict:
        output = f'Dict[{t_args[0]!s},{t_args[1]!s}]'
    elif origin is Literal:
        output = type(t_args[0]).__name__
    elif isinstance(type_, ForwardRef):
        output = type_.__forward_arg__
    elif type_ is Any:
        output = 'Any'
    else:
        output = type_.__name__

    if level == 0:
        return f'`{output!s}`{", optional" if optional else ""}'
    return output


class TweettrMeta(type):
    """meta class for TweettrBase"""
    def __new__(mcs, name, bases, attrs):
        # set slots if not defined.
        attrs.setdefault('__slots__', ())

        # add annotations in alphabetical order to docstring
        doc_str = attrs.get('__doc__', None)
        if doc_str:
            m = re.search(r"^(?P<indent>\s*)[{]attributes[}]\s*", doc_str, re.MULTILINE)
            if m:
                attribute_items = sorted(attrs.get('__annotations__', {}).items())
                doc_attrs = []
                for attr_name, attr_type in attribute_items:
                    type_str = _typing_to_np_doc(attr_type)
                    doc_attr = f'{attr_name}: {type_str}'
                    doc_attrs.append(doc_attr)
                indent = m.group('indent')
                rendered_attributes = f'\n{indent}'.join(doc_attrs)
                doc_str = doc_str.format(attributes=rendered_attributes)
                attrs['__doc__'] = doc_str

        return super().__new__(mcs, name, bases, attrs)


class TweettrBase(metaclass=TweettrMeta):
    """TweettrBase base class for attribute access to tweet dictionary

    Notes
    -----
      - getitem access provides access to all keys in the dictionary
      - attribute access requires the attribute to be type annotated
    """
    __slots__ = ('_dict', )

    def __init__(self, json_data: Dict[str, Any]):
        self._dict = json_data

    def __getattr__(self, item: str):
        if item in self.__annotations__:
            try:
                return self._dict[item]
            except KeyError:
                return None
        raise AttributeError(f"type object '{self.__class__.__name__}' has no attribute '{item}'")

    def __getitem__(self, item: str):
        return self._dict[item]


def enhance_type_hinted_classes_with_descriptors(cls):
    """enhances the provided cls with non-data descriptors

    attaches a TweettrDescriptor for each type annotated with a subclass of
    TweettrBase. This allows dynamic encapsulation into the specific TweettrBase
    subclass at runtime.

    """
    if not issubclass(cls, TweettrBase):
        raise ValueError(f'class has to be subclass of TweettrBase')
    # Note: the only reason this is required is because of ForwardRefs
    #   otherwise it would be cleaner to do this in the meta class
    type_hints = get_type_hints(cls)
    cls.__annotations__.update(type_hints)  # update with de-referenced hints

    for attr_name, attr_type in cls.__annotations__.items():

        origin = get_origin(attr_type)
        t_args = get_args(attr_type)

        optional = origin is Union and type(None) in t_args
        as_list = origin is list

        if optional or as_list:
            attr_type = t_args[0]

        if isinstance(attr_type, type) and issubclass(attr_type, TweettrBase):
            attr_desc = TweettrDescriptor(attr_name, attr_type,
                                          optional=optional, as_list=as_list)
            setattr(cls, attr_name, attr_desc)



"""obfuscate tweet data for tests"""
from datetime import datetime
import json
import time


tdict = str.maketrans("ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz",
                      "WWWWWWWWWWWWWWWWWWWWWWWWWWwwwwwwwwwwwwwwwwwwwwwwwwww")


def obfuscate(data, key=None):
    if isinstance(data, list):
        out = []
        for item in data:
            out.append(obfuscate(item))
        return out
    elif isinstance(data, dict):
        out = {}
        for key, item in data.items():
            out[key] = obfuscate(item, key=key)
        return out
    else:
        if key == "created_at":
            x = datetime.utcnow().utctimetuple()
            x = time.strftime("%a %b %d %H:%M:%S %z %Y", x)
            return x
        elif key == "id":
            return 11223344556677889900
        elif key == "id_str":
            return '11223344556677889900'
        elif key == "screen_name":
            return "A" * len(data)
        elif isinstance(data, str):
            return data.translate(tdict)

        return data


if __name__ == '__main__':

    import pprint
    import sys

    if not len(sys.argv) == 3:
        print(__file__, 'infile', 'outfile')
        sys.exit(1)

    with open(sys.argv[1], 'r') as f:
        data = json.load(f)

    new_data = obfuscate(data)

    pprint.pprint(new_data)

    with open(sys.argv[2], 'w') as f:
        json.dump(new_data, f, indent=2)



import pytest
import glob
import os.path
import json
from tweettr import Tweet, User


TWEET_FNS = list(sorted(glob.glob(os.path.join(os.path.dirname(__file__), 'tweets', '*.json'))))


@pytest.fixture(scope='function',
                params=TWEET_FNS,
                ids=[os.path.basename(fn) for fn in TWEET_FNS])
def tweet_fn(request):
    tweet_fn = request.param
    yield tweet_fn


def test_tweet_constructor(tweet_fn):
    with open(tweet_fn, 'r') as f:
        data = json.load(f)

    tweet = Tweet(data)
    # verify that the internal dict is a reference to data
    assert tweet._dict is data


def test_tweet_attribute_access(tweet_fn):
    tweet = Tweet.from_file(tweet_fn)

    assert tweet.id > 0
    assert isinstance(tweet.user, User)
    assert tweet.timestamp_ms
    assert tweet.user.id > 0
    assert tweet.entities.hashtags[0].text


def test_tweet_non_existing_attribute_access(tweet_fn):
    tweet = Tweet.from_file(tweet_fn)

    with pytest.raises(AttributeError):
        _ = tweet.i_do_not_exist


def test_tweet_getitem_access(tweet_fn):
    tweet = Tweet.from_file(tweet_fn)

    assert tweet['id_str'] == tweet.id_str


def test_tweet_nocopy_class_object_wrapping(tweet_fn):
    tweet = Tweet.from_file(tweet_fn)

    # noinspection PyProtectedMember
    assert tweet['user'] is tweet.user._dict


def test_tweet_emoji_parsing(tweet_fn):
    tweet = Tweet.from_file(tweet_fn)

    assert isinstance(tweet.emojis, list)


def test_emoji_parsing():
    # noinspection PyProtectedMember
    from tweettr._emoji import iter_emoji_entities

    txt = b'fish \xf0\x9f\x90\x8b fush'.decode('utf8')
    emojis = list(iter_emoji_entities(txt))
    assert len(emojis) == 1


def test_tweet_attributes_recursively(tweet_fn):
    # noinspection PyProtectedMember
    from tweettr._base import TweettrBase

    tweet = Tweet.from_file(tweet_fn)

    def recurse(t):
        for attr in t.__annotations__:
            v = getattr(t, attr)
            if isinstance(v, TweettrBase):
                recurse(v)
    recurse(tweet)



"""tweettr minimal tweet encapsulation

from Twitter:
  ## Parsing best-practices ##
  - Twitter JSON is encoded using UTF-8 characters.
  - Parsers should tolerate variance in the ordering of fields with ease.
    It should be assumed that Tweet JSON is served as an unordered hash of data.
  - Parsers should tolerate the addition of 'new' fields. The Twitter platform
    has continually evolved since 2006, so there is a long history of new metadata
    being added to Tweets.
  - JSON parsers must be tolerant of ‘missing’ fields, since not all fields appear
    in all contexts.
  - It is generally safe to consider a nulled field, an empty set, and the absence
    of a field as the same thing.
"""
__author__ = "Andreas Poehlmann"
__email__ = "andreas@poehlmann.io"

import json
from typing import Any, Dict, List, Literal, Optional

from ._base import TweettrBase, enhance_type_hinted_classes_with_descriptors
from ._emoji import iter_emoji_entities

__all__ = [
    'BoundingBox',
    'Coordinates',
    'Emoji',
    'Entities',
    'ExtendedEntities',
    'ExtendedTweet',
    'HashTag',
    'Media',
    'Place',
    'Poll',
    'QuotedStatusPermalink',
    'Size',
    'Sizes',
    'Symbol',
    'Tweet',
    'Url',
    'User',
    'UserMention',
]


class User(TweettrBase):
    """User
    
    see: https://developer.twitter.com/en/docs/tweets/data-dictionary/overview/user-object

    Attributes
    ----------
    {attributes}
    """
    id: int
    id_str: str
    name: str
    screen_name: str

    location: Optional[str]
    url: Optional[str]
    description: Optional[str]
    protected: bool
    verified: bool

    followers_count: int
    friends_count: int
    listed_count: int
    favourites_count: int
    statuses_count: int
    created_at: str

    default_profile: bool
    default_profile_image: bool
    profile_image_url_https: str
    profile_banner_url: Optional[str]

    withheld_in_countries: Optional[List[str]]
    withheld_scope: Optional[str]

    # deprecated
    # ----------
    # - utc_offset
    # - time_zone
    # - lang
    # - geo_enabled
    # - following
    # - follow_request_sent
    # - notifications
    # - contributors_enabled
    # - profile_image_url
    # - profile_background_color
    # - profile_background_image_url
    # - profile_background_image_url_https
    # - profile_background_tile
    # - profile_link_color
    # - profile_sidebar_border_color
    # - profile_sidebar_fill_color
    # - profile_text_color
    # - profile_use_background_image
    # - is_translator
    # - translator_type

    def __repr__(self):
        return f'{self.__class__.__name__}(screen_name={repr(self.screen_name)}, id={repr(self.id)})'


class Coordinates(TweettrBase):
    """Coordinates
    
    see: https://developer.twitter.com/en/docs/tweets/data-dictionary/overview/geo-objects#coordinates

    Attributes
    ----------
    {attributes}
    """
    coordinates: List[float]
    type: Literal['Point']

    def __repr__(self):
        long, lat = self.coordinates
        return f'{self.__class__.__name__}(longitude={repr(long)}, latitude={repr(lat)})'


class BoundingBox(TweettrBase):
    """BoundingBox
    
    see: https://developer.twitter.com/en/docs/tweets/data-dictionary/overview/geo-objects#obj-boundingbox

    Attributes
    ----------
    {attributes}
    """
    coordinates: List[List[List[float]]]
    type: Literal['Polygon']

    def __repr__(self):
        num_bb = len(self.coordinates)
        return f'{self.__class__.__name__}(num_bb={num_bb})'


class Place(TweettrBase):
    """Place

    see: https://developer.twitter.com/en/docs/tweets/data-dictionary/overview/geo-objects#place-dictionary

    Attributes
    ----------
    {attributes}
    """
    id: str
    country: Optional[str]
    country_code: Optional[str]
    full_name: Optional[str]
    name: Optional[str]
    place_type: Optional[str]
    url: Optional[str]
    attributes: Optional[dict]

    bounding_box: Optional[BoundingBox]

    def __repr__(self):
        return f'{self.__class__.__name__}(name={repr(self.name)}, id={repr(self.id)})'


class HashTag(TweettrBase):
    """HashTag
    
    see: https://developer.twitter.com/en/docs/tweets/data-dictionary/overview/entities-object#hashtags

    Attributes
    ----------
    {attributes}
    """
    text: str
    indices: List[int]

    def __repr__(self):
        return f'{self.__class__.__name__}({repr(self.text)})'


class Emoji(TweettrBase):
    """Emoji

    this is a custom Tweettr object that acts like a HashTag object but contains info about
    emojis in the tweet text

    Attributes
    ----------
    {attributes}
    """
    text: str
    indices: List[int]

    def __repr__(self):
        return f'{self.__class__.__name__}({repr(self.text)})'


class Size(TweettrBase):
    """Size
    
    see: https://developer.twitter.com/en/docs/tweets/data-dictionary/overview/entities-object#size

    Attributes
    ----------
    {attributes}
    """
    w: int
    h: int
    resize: str

    def __repr__(self):
        return f'{self.__class__.__name__}(size={self.w}x{self.h}, resize={repr(self.resize)})'


class Sizes(TweettrBase):
    """Sizes

    simple container for multiple Size objects

    see: https://developer.twitter.com/en/docs/tweets/data-dictionary/overview/entities-object#size

    Attributes
    ----------
    {attributes}
    """
    large: Size
    medium: Size
    small: Size
    thumb: Size

    def __repr__(self):
        return f'{self.__class__.__name__}(instance at {id(self)})'


class Symbol(TweettrBase):
    """Symbol
    
    see: https://developer.twitter.com/en/docs/tweets/data-dictionary/overview/entities-object#symbols

    Attributes
    ----------
    {attributes}
    """
    text: str
    indices: List[int]

    def __repr__(self):
        return f'{self.__class__.__name__}({repr(self.text)})'


class Url(TweettrBase):
    """Url

    see: https://developer.twitter.com/en/docs/tweets/data-dictionary/overview/entities-object#urls

    Attributes
    ----------
    {attributes}
    """
    display_url: str
    expanded_url: str
    url: str
    indices: List[int]

    def __repr__(self):
        return f'{self.__class__.__name__}({repr(self.url)})'


class Media(TweettrBase):
    """Media
    
    see: https://developer.twitter.com/en/docs/tweets/data-dictionary/overview/entities-object#media

    Attributes
    ----------
    {attributes}
    """
    display_url: str
    expanded_url: str
    id: int
    id_str: str
    media_url: str
    media_url_https: str
    sizes: Sizes

    source_status_id: Optional[int]
    source_status_id_str: Optional[str]

    indices: List[int]
    type: str
    url: str

    # not used ?
    # ----------
    # source_user_id
    # source_user_id_str
    # description

    def __repr__(self):
        return f'{self.__class__.__name__}(id={repr(self.id)})'


class UserMention(TweettrBase):
    """UserMention
    
    see: https://developer.twitter.com/en/docs/tweets/data-dictionary/overview/entities-object#mentions

    Attributes
    ----------
    {attributes}
    """
    id: int
    id_str: str
    indices: List[int]
    name: str
    screen_name: str

    def __repr__(self):
        return f'{self.__class__.__name__}(mention={repr(self.screen_name)})'


class Poll(TweettrBase):
    """Poll

    see: https://developer.twitter.com/en/docs/tweets/data-dictionary/overview/entities-object#polls

    Attributes
    ----------
    {attributes}
    """
    options: List[Dict]  # TODO: should be List[Option]
    end_datetime: str
    duration_minutes: str

    def __repr__(self):
        num_options = len(self.options)
        return f'{self.__class__.__name__}(num_options={num_options})'


class Entities(TweettrBase):
    """Entities
    
    see: https://developer.twitter.com/en/docs/tweets/data-dictionary/overview/entities-object

    Attributes
    ----------
    {attributes}
    """
    hashtags: List[HashTag]
    urls: List[Url]
    user_mentions: List[UserMention]
    symbols: List[Symbol]
    polls: List[Poll]
    media: List[Media]

    def __repr__(self):
        return f'{self.__class__.__name__}(instance at {id(self)})'


class ExtendedEntities(TweettrBase):
    """ExtendedEntities
    
    see: https://developer.twitter.com/en/docs/tweets/data-dictionary/overview/extended-entities-object

    Attributes
    ----------
    {attributes}
    """
    media: List[Media]

    def __repr__(self):
        return f'{self.__class__.__name__}(instance at {id(self)})'


class ExtendedTweet(TweettrBase):
    """ExtendedTweet

    this depends on api-access and streaming or not endpoints and if extended_mode is requested
    
    see: https://developer.twitter.com/en/docs/tweets/tweet-updates

    Attributes
    ----------
    {attributes}
    """
    full_text: str
    truncated: Literal[False]
    entities: Optional[Entities]
    extended_entities: Optional[ExtendedEntities]
    display_text_range: Optional[List[int]]

    def __repr__(self):
        return f'{self.__class__.__name__}(instance at {id(self)})'


class QuotedStatusPermalink(TweettrBase):
    """QuotedStatusPermalink
    
    see: https://developer.twitter.com/en/docs/tweets/data-dictionary/overview/intro-to-tweet-json

    Attributes
    ----------
    {attributes}
    """
    display: str
    expanded: str
    url: str

    def __repr__(self):
        return f'{self.__class__.__name__}(display={repr(self.display)})'


class Tweet(TweettrBase):
    """Tweet

    see: https://developer.twitter.com/en/docs/tweets/data-dictionary/overview/tweet-object

    Attributes
    ----------
    {attributes}
    """
    created_at: str
    id: int
    id_str: str
    text: str
    source: str
    truncated: bool

    contributors: Any

    user: User

    is_quote_status: bool
    quote_count: int

    reply_count: int
    retweet_count: int
    favorite_count: int

    favorited: bool
    retweeted: bool

    filter_level: str
    lang: str
    timestamp_ms: Optional[str]

    possibly_sensitive: Optional[bool]
    quoted_status_id: Optional[int]
    quoted_status_id_str: Optional[str]
    in_reply_to_status_id: Optional[int]
    in_reply_to_status_id_str: Optional[str]
    in_reply_to_user_id: Optional[int]
    in_reply_to_user_id_str: Optional[str]
    in_reply_to_screen_name: Optional[str]

    coordinates: Optional[Coordinates]
    place: Optional[Place]

    display_text_range: Optional[List[int]]
    retweeted_status: Optional['Tweet']
    quoted_status: Optional['Tweet']
    entities: Optional[Entities]
    extended_entities: Optional[ExtendedEntities]
    extended_tweet: Optional[ExtendedTweet]
    quoted_status_permalink: Optional[QuotedStatusPermalink]

    # deprecated
    # ----------
    # - geo

    @classmethod
    def from_file(cls, filename):
        with open(filename, 'r') as json_f:
            data = json.load(json_f)
        return cls(data)

    def __repr__(self):
        try:
            screen_name = self.user.screen_name
        except (AttributeError, TypeError):
            screen_name = None
        return f'{self.__class__.__name__}(id={repr(self.id)}, user={repr(screen_name)})'

    @property
    def permalink(self):
        """return a permalink to the tweet"""
        return f'https://twitter.com/{self.user.screen_name}/status/{self.id}'

    @property
    def is_retweet(self):
        return True if self.retweeted_status else False

    @property
    def is_quote_tweet(self):
        return True if self.quoted_status else False

    @property
    def is_original_tweet(self):
        return not (self.is_retweet or self.is_quote_tweet)

    @property
    def extended_full_text(self):
        """return the untruncated full_text"""
        return self.extended_tweet.full_text if self.extended_tweet else self.text

    @property
    def emojis(self):
        """return a list of all Emoji Entities in tweet"""
        return [Emoji(data) for data in iter_emoji_entities(self.extended_full_text)]


# Set descriptors for all type annotations that contain subclasses
# of TweettrBase to allow transparent conversion of sub dicts to tweettr classes
for subclass in TweettrBase.__subclasses__():
    # Note: the only way to do this in a meta class is if ForwardRefs would
    #   be dynamically resolved lazily during runtime...
    #   Let's prevent that overhead and do it after all subclasses have been defined.
    #   Nice side effect: this TypeErrors if a forward ref can't be de-referenced
    enhance_type_hinted_classes_with_descriptors(subclass)



"""tweettr

Thin `attrdict` like wrapper around (a few) tweet objects
"""
__author__ = "Andreas Poehlmann"
__email__ = "andreas@poehlmann.io"

from ._objects import (BoundingBox, Coordinates, Emoji, Entities,
                       ExtendedEntities, ExtendedTweet, HashTag, Media,
                       Place, Poll, QuotedStatusPermalink, Size, Sizes,
                       Symbol, Tweet, Url, User, UserMention)
try:
    from ._version import __version__
except ImportError:
    __version__ = None



"""
tweettr
=======

Very thin wrapper around (a few) twitter objects

Author: Andreas Poehlmann
Email: andreas@poehlmann.io
"""
from setuptools import setup, find_packages


setup(
    name='tweettr',
    author='Andreas Poehlmann',
    author_email='andreas@poehlmann.io',
    url='https://github.com/ap--/tweettr',
    license='MIT',
    use_scm_version={
        'write_to': 'src/tweettr/_version.py',
        'write_to_template': '__version__ = "{version}"',
        'tag_regex': r'^(?P<prefix>v)?(?P<version>[^\+]+)(?P<suffix>.*)?$',
    },
    setup_requires=[
        'setuptools_scm',
        'pytest-runner'
    ],
    python_requires='>=3.8',
    tests_require=[
        'pytest',
    ],
    packages=find_packages(where='src'),
    package_dir={
        '': 'src'
    },
    description='tweettr is a `attrdict` like wrapper around (a few) Twitter JSON objects.',
    long_description=open('README.md').read(),
    long_description_content_type="text/markdown",
    classifiers=[
        'Programming Language :: Python :: 3.8',
    ]
)
